"""
Selectors improve ``.loc[]`` indexing for multi-index pandas data.
"""

from collections.abc import Mapping
from functools import reduce
from operator import and_
from typing import Any, Optional, Union

import numpy as np
from attrs import define, field
from pandas import DataFrame, Index, MultiIndex, Series

from .core import ensure_multiindex, projectlevel
from .types import Data
from .utils import shell_pattern_to_regex


def maybe_const(x):
    return x if isinstance(x, Selector) else Const(x)


class Selector:
    # Tell numpy and pandas that we want precedence
    __array_ufunc__ = None
    __pandas_priority__ = 5000

    def __invert__(self):
        return Not(self)

    def __and__(self, other):
        if isinstance(other, Special):
            return other & self
        return And(self, maybe_const(other))

    def __or__(self, other):
        if isinstance(other, Special):
            return other | self
        return Or(self, maybe_const(other))

    __rand__ = __and__
    __ror__ = __or__


@define
class BinOp(Selector):
    a: Selector
    b: Selector


@define
class Const(Selector):
    val: Any

    def __call__(self, df):
        return self.val(df) if callable(self.val) else self.val


@define
class And(BinOp):
    def __call__(self, df):
        return self.a(df) & self.b(df)


@define
class Or(BinOp):
    def __call__(self, df):
        return self.a(df) | self.b(df)


@define
class Not(Selector):
    a: Selector

    def __call__(self, df):
        return ~self.a(df)


class Special(Selector):
    pass


@define
class AllSelector(Special):
    def __invert__(self):
        return NoneSelector()

    def __and__(self, other):
        return other

    def __or__(self, other):
        return self

    def __call__(self, df):
        return Series(True, df.index)


@define
class NoneSelector(Special):
    def __invert__(self):
        return AllSelector()

    def __and__(self, other):
        return self

    def __or__(self, other):
        return other

    __rand__ = __and__
    __ror__ = __or__

    def __call__(self, df):
        return Series(False, df.index)


# Singletons for easy access
All = AllSelector()
None_ = NoneSelector()


@define
class Isin(Selector):
    filters: Mapping[str, Any]
    ignore_missing_levels: bool = field(default=False, repr=False)

    def __call__(self, df):
        if isinstance(df, Index):
            index = df
        else:
            index = df.index

        filters = self.filters
        if self.ignore_missing_levels:
            filters = {k: v for k, v in filters.items() if k in index.names}

        def apply_filter(value, level):
            if callable(value):
                return value(index.get_level_values(level))
            return index.isin(np.atleast_1d(value), level=level)

        return Series(
            reduce(and_, (apply_filter(v, k) for k, v in filters.items())), index
        )


@define
class IsinIndex(Selector):
    index: MultiIndex
    ignore_missing_levels: bool = field(default=False, repr=False)

    def __call__(self, df):
        if isinstance(df, Index):
            index = df
        else:
            index = df.index

        missing_levels = self.index.names.difference(index.names)
        if missing_levels:
            if not self.ignore_missing_levels:
                raise KeyError(
                    f"selecting on levels {self.index.names}, but only given: {index.names}"
                )
            sel_index = self.index.droplevel(list(missing_levels))
        else:
            sel_index = self.index

        return Series(projectlevel(index, sel_index.names).isin(sel_index), index)


def isin(
    df: Union[Data, Index, None] = None,
    index: Optional[Index] = None,
    /,
    ignore_missing_levels: bool = False,
    **filters: Any,
) -> Union[Isin, IsinIndex, Series]:
    """
    Constructs a MultiIndex selector.

    Parameters
    ----------
    df : Data, optional
        Data on which to match, if missing an ``Isin`` object is returned
    index : Index, optional
        Filter based on common levels given in index. Can also be passed as the
        ``df`` argument. Cannot be combined with filters.
    ignore_missing_levels : bool, default False
        If set, levels missing in data index will be ignored
    **filters
        Filter to apply on given levels (lists are ``or`` ed, levels are ``and`` ed)
        Callables are evaluated on the index level values.

    Returns
    -------
    Isin, IsinIndex or Series

    Example
    -------
    >>> df.loc[isin(region="World", gas=["CO2", "N2O"])]

    or with explicit df to get a boolean mask

    >>> isin(df, region="World", gas=["CO2", "N2O"])

    For selecting across multiple levels, a multiindex is passed as positional argument:

    >>> index = pd.MultiIndex.from_tuples(
    ...     [("World", "CO2"), ("R10_EUROPE", "N2O")],
    ...     names=["region", "gas"],
    ... )
    >>> df.loc[isin(index)]

    """
    if not filters and (index is not None or isinstance(df, Index)):
        if index is None and isinstance(df, Index):
            # Special case: index argument was passed into df
            index, df = df, None

        tester = IsinIndex(
            ensure_multiindex(index), ignore_missing_levels=ignore_missing_levels
        )
    else:
        tester = Isin(filters, ignore_missing_levels=ignore_missing_levels)

    return tester if df is None else tester(df)


@define
class Ismatch(Selector):
    filters: Mapping[str, Any]
    regex: bool = False
    ignore_missing_levels: bool = field(default=False, repr=False)

    def index_match(self, index, patterns):
        matches = np.zeros((len(index),), dtype=bool)
        for pat in patterns:
            if isinstance(pat, str):
                if not self.regex:
                    pat = shell_pattern_to_regex(pat) + "$"
                matches |= index.str.match(pat, na=False)
            else:
                matches |= index == pat
        return matches

    def multiindex_match(self, index, patterns, level):
        level_num = index.names.index(level)
        levels = index.levels[level_num]

        matches = self.index_match(levels, patterns)

        (indices,) = np.where(matches)
        return np.isin(index.codes[level_num], indices)

    def __call__(self, df):
        if isinstance(df, Index):
            index = df
        else:
            index = df.index

        single = self.filters.get(None) if len(self.filters) == 1 else None
        if single is not None:
            matches = self.index_match(index, np.atleast_1d(single))
        else:
            filters = self.filters
            if self.ignore_missing_levels:
                filters = {k: v for k, v in filters.items() if k in index.names}

            tests = (
                self.multiindex_match(index, np.atleast_1d(v), level=k)
                for k, v in filters.items()
            )
            matches = reduce(and_, tests)

        return Series(matches, index=index)


def ismatch(
    df: Union[None, Index, DataFrame, Series, str] = None,
    singlefilter: Optional[str] = None,
    /,
    regex: bool = False,
    ignore_missing_levels: bool = False,
    **filters,
) -> Union[Ismatch, Series]:
    """
    Constructs an Index or MultiIndex selector based on pattern matching.

    Parameters
    ----------
    df : Data, optional
        Data on which to match, if missing an ``Isin`` object is returned.
    singlefilter : str, optional
        Filter to apply on a non-multiindex index (can also be handed into the ``df``
        argument)
    regex : bool, default False
        If set, filters are interpreted as plain regex strings, otherwise (by default) a
        glob-like syntax is used
    ignore_missing_levels : bool, default False
        If set, levels missing in data index will be ignored
    **filters
        Filter to apply on given levels (lists are ``or`` ed, levels are ``and`` ed)

    Returns
    -------
    Isin or Series

    Example
    -------
    for a multiindex:

    >>> df.loc[ismatch(variable="Emissions|*|Fossil Fuel and Industry")]

    for a single index:

    >>> df.loc[ismatch("*bla*")]

    """
    if not filters and singlefilter is not None:
        filters = {None: singlefilter}

    if df is None:
        return Ismatch(
            filters, regex=regex, ignore_missing_levels=ignore_missing_levels
        )
    elif not isinstance(df, (DataFrame, Series, Index)):
        # Special case: a pattern was passed in through `df` which wants to be applied to
        # hopefully a non-MultiIndex based Series or dataframe that we get afterwards
        filters = {None: df}
        return Ismatch(filters, regex=regex)
    else:
        return Ismatch(
            filters, regex=regex, ignore_missing_levels=ignore_missing_levels
        )(df)



"""
Unit handling in pandas data.

Enables unit conversions based on pint's application registry (see also Notes).

By default units are expected -- as in the IAMC default format -- on a ``unit`` level on
each row, but a column-wise ``unit`` level is also supported.

Units can be handled in one of two flavours:

1. :py:func:`convert_unit` converts manually to a new unit like `convert_unit(s, "km")`

2. :py:func:`quantify` convert data to a pint pandas array which tracks units implicitly
   through arithmetics until :py:func:`dequantify` then extracts the tracked unit back
   into the multiindex level.

   While this is in theory the simpler approach, the underlying library ``pint-pandas`` [1]_
   is brittle and breaks from time to time.

Notes
-----
The pint application registry is set by :py:func:`pint.set_application_registry` or
with :py:func:`set_openscm_registry_as_default`. The latter sets the IAMC based ``openscm-units`` one [2]_.

Examples
--------
>>> import pandas_indexing as pi
>>> pi.set_openscm_registry_as_default()
>>> s = Series(
...     [7, 8],
...     MultiIndex.from_tuples([("foo", "mm"), ("bar", "m")], names=["var", "unit"]),
... )
>>> s = pi.convert_unit(s, "km")
>>> s
var  unit
bar  km      0.008000
foo  km      0.000007
dtype: float64

>>> pi.quantify(s)
var
bar    0.008
foo    7e-06
dtype: pint[kilometer]

References
----------
.. [1] https://github.com/hgrecco/pint-pandas
.. [2] https://github.com/openscm/openscm-units

See Also
--------
pint.set_application_registry

"""

from collections.abc import Mapping
from typing import Callable, Optional, Union

from pandas import DataFrame, Series

from .core import assignlevel, uniquelevel
from .types import Axis, Data
from .utils import LazyLoader, doc


INSTALL_PACKAGE_WARNING = (
    "Please install {name} via conda or pip, or use the pandas-indexing[units] variant."
)

pint_pandas = LazyLoader(
    "pint_pandas",
    globals(),
    "pint_pandas",
    INSTALL_PACKAGE_WARNING.format(name="pint_pandas"),
)
pint = LazyLoader(
    "pint", globals(), "pint", INSTALL_PACKAGE_WARNING.format(name="pint")
)


@doc(
    data="""
    data : DataFrame or Series
        DataFrame or Series to quantify
    """,
    example_call="quantify(s)",
)
def quantify(
    data: Data,
    level: str = "unit",
    unit: Optional[str] = None,
    axis: Axis = 0,
    copy: bool = False,
) -> Data:
    """
    Convert columns in `data` to pint extension types to handle units.

    `pint-pandas <https://github.com/grecco/pint-pandas>`_ can only represent a single
    unit per column and is somewhat brittle.

    Parameters
    ----------{data}
    unit : str, optional
        If given, assumes data is currently in this unit.
    level : str, optional
        Level of which to use the unit, by default "unit"
    axis : Axis, optional
        Axis from which to pop the `level`, by default 0
    copy : bool, optional
        Whether data should be copied, by default False

    Returns
    -------
    Data
        Data with internalized unit which stays with arithmetics

    Raises
    ------
    ValueError
        If `level` contains more than one unit

    Examples
    --------
    >>> s = Series(
    ...     [7e-3, 8],
    ...     MultiIndex.from_tuples([("foo", "m"), ("bar", "m")], names=["var", "unit"]),
    ... )
    >>> {example_call}
    var
    foo    7e-06
    bar    0.008
    dtype: pint[kilometer]

    Notes
    -----
    pint-pandas uses the pint application registry, which can be set with
    :py:func:`pint.set_application_registry` or
    :py:func:`set_openscm_registry_as_default`.

    See Also
    --------
    set_openscm_registry_as_default
    dequantify
    convert_unit

    """
    if unit is None:
        unit = uniquelevel(data, level, axis)
        if len(unit) != 1:
            raise ValueError(
                f"pint-pandas can only represent homogeneous units: {', '.join(unit)}"
            )
        unit = unit[0]
        data = data.droplevel(level, axis=axis)

    return data.astype(pint_pandas.PintType(unit), copy=copy)


def format_dtype(dtype):
    if not isinstance(dtype, pint_pandas.PintType):
        return ""

    return ("{:" + dtype.ureg.default_format + "}").format(dtype.units)


def dequantify(data: Data, level: str = "unit", axis: Axis = 0, copy: bool = False):
    if isinstance(data, Series):
        unit = format_dtype(data.dtype)
        data = data.pint.magnitude
    elif isinstance(data, DataFrame):
        if axis in (0, "index"):
            unit = data.dtypes.unique()
            if len(unit) != 1:
                raise ValueError(
                    f"Only homogeneous units can be represented: {', '.join(unit)}"
                )
            unit = format_dtype(unit[0])
            data = data.apply(lambda s: s.pint.magnitude, result_type="expand", axis=1)
        elif axis in (1, "columns"):
            unit = data.dtypes.map(format_dtype)
            data = data.apply(lambda s: s.pint.magnitude, result_type="expand", axis=0)
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )
    else:
        raise TypeError(
            f"data must derive from DataFrame or Series, but is {type(data)}"
        )

    return assignlevel(data, unit=unit, axis=axis)


@doc(
    data="""
    data : DataFrame or Series
        DataFrame or Series with a "unit" level
    """,
    convert_unit_s_km='convert_unit(s, "km")',
    convert_unit_s_m_to_km='convert_unit(s, {"m": "km"})',
)
def convert_unit(
    data: Data,
    unit: Union[str, Mapping[str, str], Callable[[str], str]],
    level: Optional[str] = "unit",
    axis: Axis = 0,
):
    """
    Converts units in a dataframe or series.

    Parameters
    ----------{data}
    unit : str or dict or function from old to new unit
        Either a single target unit or a mapping from old unit to target unit
        (a unit missing from the mapping or with a return value of None is kept)
    level : str|None, default "unit"
        Level name on ``axis``
        If None, then ``unit`` needs to be a mapping like ``{{from_unit: to_unit}}``
    axis : Axis, default 0
        Axis of unit level

    Returns
    -------
    Data
        DataFrame or Series with converted units

    Examples
    --------
    >>> s = Series(
    ...     [7, 8],
    ...     MultiIndex.from_tuples(
    ...         [("foo", "mm"), ("bar", "m")], names=["var", "unit"]
    ...     ),
    ... )
    >>> {convert_unit_s_km}
    var  unit
    bar  km      0.008000
    foo  km      0.000007
    dtype: float64

    >>> {convert_unit_s_m_to_km}
    var  unit
    bar  km      0.008
    foo  mm      7.000
    dtype: float64

    Notes
    -----
    Uses the pint application registry, which can be set with
    :py:func:`pint.set_application_registry` or
    :py:func:`set_openscm_registry_as_default`.

    See Also
    --------
    set_openscm_registry_as_default
    quantify
    dequantify

    """
    if callable(unit):
        unit_map = unit
    elif isinstance(unit, dict):
        unit_map = unit.get
    else:
        unit_map = lambda u: unit

    ur = pint.get_application_registry()

    def _convert_unit(df, old_unit=None):
        if old_unit is None:
            old_unit = df.name
        new_unit = unit_map(old_unit)
        if new_unit is None:
            return df

        factor = ur.Quantity(1, old_unit).to(new_unit).m
        df = factor * df
        if level is None:
            return df
        return assignlevel(df, **{level: new_unit})

    try:
        if level is None:
            if not (isinstance(unit, dict) and len(unit) == 1):
                raise ValueError(
                    "If level is None, unit must look like {{fromunit: tounit}}"
                )
            (old_unit,) = unit.keys()
            data = _convert_unit(data, old_unit=old_unit)
        elif axis in (0, "index"):
            data = data.groupby(level=level, group_keys=False).apply(_convert_unit)
        elif axis in (1, "columns"):
            data = data.T.groupby(level=level, group_keys=False).apply(_convert_unit).T
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )

        return data.__finalize__(data, "convert_unit")
    except pint.errors.DimensionalityError as exc:
        raise exc from None  # remove exception chaining


_openscm_registry = None


def get_openscm_registry(add_co2e: bool = True):
    global _openscm_registry
    if _openscm_registry is not None:
        return _openscm_registry

    import openscm_units

    if add_co2e:
        _openscm_registry = openscm_units.ScmUnitRegistry()
        _openscm_registry.add_standards()
        _openscm_registry.define("CO2e = CO2")
        _openscm_registry.define("CO2eq = CO2")
    else:
        _openscm_registry = openscm_units.unit_registry

    return _openscm_registry


def set_openscm_registry_as_default(add_co2e: bool = True):
    unit_registry = get_openscm_registry(add_co2e=add_co2e)

    app_registry = pint.get_application_registry()
    if unit_registry is not app_registry:
        pint.set_application_registry(unit_registry)

    return unit_registry


def is_unit(unit: str) -> bool:
    ur = pint.get_application_registry()
    try:
        ur.Unit(unit)
        return True
    except pint.errors.UndefinedUnitError:
        return False



"""
Utils module.

Simple utility functions not of greater interest
"""

import importlib
import re
from textwrap import dedent
from types import ModuleType
from typing import Any, Union

from attrs import define
from pandas import DataFrame, Index, Series
from pandas.util._decorators import doc as pandadoc

from .types import Axis


@pandadoc(pandadoc)
def doc(*docstrings, **params):
    return pandadoc(*docstrings, **{p: dedent(v).rstrip() for p, v in params.items()})


def shell_pattern_to_regex(s):
    """
    Escape characters with specific regexp use.
    """
    return re.escape(s).replace(r"\*\*", ".*").replace(r"\*", r"[^|]*")


def print_list(x, n):
    """
    Return a printable string of a list shortened to n characters.

    Copied from pyam.utils.print_list by Daniel Huppmann, licensed under
    Apache 2.0.

    https://github.com/IAMconsortium/pyam/blob/449b77cb1c625b455e3675801477f19e99b30e64/pyam/utils.py#L599-L638
    .
    """
    # if list is empty, only write count
    if len(x) == 0:
        return "(0)"

    # write number of elements, subtract count added at end from line width
    x = [i if i != "" else "''" for i in map(str, x)]
    count = f" ({len(x)})"
    n -= len(count)

    # if not enough space to write first item, write shortest sensible line
    if len(x[0]) > n - 5:
        return "..." + count

    # if only one item in list
    if len(x) == 1:
        return f"{x[0]} (1)"

    # add first item
    lst = f"{x[0]}, "
    n -= len(lst)

    # if possible, add last item before number of elements
    if len(x[-1]) + 4 > n:
        return lst + "..." + count
    else:
        count = f"{x[-1]}{count}"
        n -= len({x[-1]}) + 3

    # iterate over remaining entries until line is full
    for i in x[1:-1]:
        if len(i) + 6 <= n:
            lst += f"{i}, "
            n -= len(i) + 2
        else:
            lst += "... "
            break

    return lst + count


def get_axis(data: Union[Index, Series, DataFrame], axis: Axis = 0):
    """
    Get axis `axis` from `data`.
    """
    if isinstance(data, Index):
        return data
    elif isinstance(data, Series):
        return data.index
    elif isinstance(data, DataFrame):
        if axis in (0, "index"):
            return data.index
        elif axis in (1, "columns"):
            return data.columns
        else:
            raise ValueError(
                f"axis can only be one of 0, 1, 'index' or 'columns', not: {axis}"
            )
    else:
        raise ValueError(
            f"data needs to be a pandas Series or DataFrame, not: {type(data)}"
        )


def quote_list(l):
    return ", ".join(f'"{s}"' for s in l)


def s(l):
    return "s" if len(l) > 1 else ""


class LazyLoader(ModuleType):
    """
    Lazily import a module, mainly to avoid pulling in large dependencies.

    `contrib`, and `ffmpeg` are examples of modules that are large and not always
    needed, and this allows them to only be loaded when they are used.

    Copied from tensorflow's agents.tf_agents.utils.lazy_loader by The TF-Agents Authors (2020), licensed under Apache 2.0
    https://github.com/tensorflow/agents/blob/737d758452990dc3c81b8aeab1a6ae4f63afa12c/tf_agents/utils/lazy_loader.py#L28-L68
    """

    def __init__(self, local_name, parent_module_globals, name, importerrormsg=None):
        self._local_name = local_name
        self._parent_module_globals = parent_module_globals
        self._importerrormsg = importerrormsg

        super().__init__(name)

    def _load(self):
        """
        Load the module and insert it into the parent's globals.
        """
        # Import the target module and insert it into the parent's namespace
        try:
            module = importlib.import_module(self.__name__)
        except ImportError:
            if self._importerrormsg is not None:
                raise ImportError(self._importerrormsg) from None
            else:
                raise

        self._parent_module_globals[self._local_name] = module

        # Update this object's dict so that if someone keeps a reference to the
        #   LazyLoader, lookups are efficient (__getattr__ is only called on lookups
        #   that fail).
        self.__dict__.update(module.__dict__)

        return module

    def __getattr__(self, item):
        module = self._load()
        return getattr(module, item)

    def __dir__(self):
        module = self._load()
        return dir(module)


@define
class EqualIdentity:
    __array_ufunc__ = None
    __pandas_priority__ = 5000

    obj: Any

    def __eq__(self, other):
        return self.obj is other



from __future__ import annotations

import operator
import re
from collections.abc import Iterator, Sequence
from functools import reduce
from itertools import product
from typing import Any, Callable, TypeVar

from attrs import define, evolve, field
from pandas import DataFrame, Index, MultiIndex, Series

from .. import arithmetics
from ..core import concat
from ..selectors import isin, ismatch
from ..utils import print_list, shell_pattern_to_regex


def _summarize(index: MultiIndex, names: Sequence[str]) -> DataFrame:
    """
    Summarize unique level values grouped by `names` levels.

    Parameters
    ----------
    index : MultiIndex
        Index to summarize
    names : Sequence[str]
        Levels to group by values by

    Returns
    -------
    DataFrame
        Summary frame

    """
    return (
        index.to_frame()
        .pix.project(names)[index.names.difference(names)]  # type: ignore
        .groupby(names)
        .agg(lambda x: print_list(set(x), n=42))
    )


def maybe_parens(provenance: str) -> str:
    if any(op in provenance for op in [" + ", " - ", " * ", " / "]):
        return f"({provenance})"
    return provenance


@define
class SharedTrigger:
    active: int = field(default=0, converter=int)

    def __enter__(self):
        self.active += 1
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.active -= 1
        return False

    def __bool__(self):
        return bool(self.active)


@define
class Context:
    """
    Context shared between all Vars instances in a Resolver.

    Notes
    -----
    Not to be instantiated by the user herself

    """

    level: str
    full_index: MultiIndex
    columns: Index
    index: list[str]
    optional_combinations: SharedTrigger = field(factory=SharedTrigger)


@define
class Var:
    """
    Instance for a single variant.

    Attributes
    ----------
    data : DataFrame
        Calculated data
    provenance : str
        Formula for how the variant was calculated

    Notes
    -----
    User does not interact with individual `Var` instances, instead she only ever holds
    `Vars` instances.

    """

    data: DataFrame
    provenance: str

    SV = TypeVar("SV", bound="Var")

    def __repr__(self) -> str:
        return f"Var {self.provenance}\n{self.data.pix}"

    @property
    def empty(self) -> bool:
        return self.data.empty

    def index(self, levels: Sequence[str]) -> MultiIndex:
        if not set(levels).issubset(self.data.index.names):
            return MultiIndex.from_tuples([], names=levels)
        return self.data.pix.unique(levels)  # type: ignore

    def _binop(
        self: SV,
        op,
        x: SV,
        y: SV,
        provenance_maker: Callable[[str, str], str],
    ) -> SV:
        if not all(isinstance(v, Var) for v in (x, y)):
            return NotImplemented
        provenance = provenance_maker(x.provenance, y.provenance)
        return self.__class__(op(x.data, y.data, join="inner"), provenance)

    def __add__(self: SV, other: SV) -> SV:
        return self._binop(arithmetics.add, self, other, lambda x, y: f"{x} + {y}")  # type: ignore

    def __sub__(self: SV, other: SV) -> SV:
        return self._binop(
            arithmetics.sub,  # type: ignore
            self,
            other,
            lambda x, y: f"{x} - {maybe_parens(y)}",
        )

    def __mul__(self: SV, other: SV) -> SV:
        return self._binop(
            arithmetics.mul,  # type: ignore
            self,
            other,
            lambda x, y: f"{maybe_parens(x)} * {maybe_parens(y)}",
        )

    def __truediv__(self: SV, other: SV) -> SV:
        return self._binop(
            arithmetics.div,  # type: ignore
            self,
            other,
            lambda x, y: f"{maybe_parens(x)} / {maybe_parens(y)}",
        )

    def __neg__(self: SV) -> SV:
        return self.__class__(-self.data, "- " + maybe_parens(self.provenance))

    def as_df(self, **assign: str) -> DataFrame:
        return self.data.pix.assign(**(assign | dict(provenance=self.provenance)))


@define
class Vars:
    """
    `Vars` holds several derivations of a variable from data in a `Resolver`

    Attributes:
    ----------
    data : list of Var
        Disjunct derivations of a single variable
    context : Context
        Shared context from the resolver
    index : MultiIndex for which any derivation has data

    Notes:
    -----
    `Vars` are created with a Resolver and are to be composed with one another.

    Example:
    -------
    >>> r = Resolver.from_data(co2emissions, level="sector")
    >>> energy = r["Energy"] | (r["Energy|Supply"] + r["Energy|Demand"])
    >>> r["Energy and Industrial Processes"] | (energy + r["Industrial Processes"])

    """

    data: list[Var]
    context: Context

    SV = TypeVar("SV", bound="Vars")
    T = TypeVar("T", bound=Index | DataFrame | Series)

    @classmethod
    def from_data(
        cls: type[SV],
        data: DataFrame,
        value: str,
        *,
        context: Context,
        provenance: str | None = None,
    ) -> SV:
        if provenance is None:
            provenance = value
        data = data.loc[isin(**{context.level: value})].droplevel(context.level)  # type: ignore
        return cls([Var(data, provenance)] if not data.empty else [], context)

    @classmethod
    def from_additionalresolver(
        cls: type[SV],
        vars: SV,
        prefix: str,
        *,
        context: Context,
    ) -> SV:
        data = [evolve(var, provenance=f"{prefix}({var.provenance})") for var in vars]
        return cls(data, evolve(context, full_index=vars.context.full_index))

    def __repr__(self) -> str:
        index = self.index
        incomplete = not self._missing(index).empty

        if self.empty:
            return "Vars empty"

        s = (
            f"Vars for {len(index)}{'*' if incomplete else ''} scenarios:\n"
            + "\n".join(
                (
                    f"* {v.provenance} ("
                    f"{len(v.index(self.context.index))}"
                    f"{'*' if not self._missing(v).empty else ''})"
                )
                for v in self.data
            )
        )

        if len(self) == 1:
            s += f"\n\nDetails (since only a single variant):\n{self.data[0]}"
        return s

    @property
    def empty(self) -> bool:
        return not self

    @property
    def index(self) -> MultiIndex:
        if self.empty:
            return MultiIndex.from_tuples([], names=self.context.index)

        return concat(v.index(self.context.index) for v in self.data).unique()

    def __bool__(self) -> bool:
        return bool(self.data)

    def __len__(self) -> int:
        return len(self.data)

    def __iter__(self) -> Iterator[Var]:
        return iter(self.data)

    def __getitem__(self: SV, i: int) -> SV:
        ret = self.data[i]
        if isinstance(ret, Var):
            ret = [ret]
        return self.__class__(ret, self.context)

    class _LocIndexer:
        def __init__(self, obj):
            self._obj = obj

        def __getitem__(self, x: Any) -> Vars:
            obj = self._obj
            return obj.__class__(
                [
                    var.__class__(z, var.provenance)
                    for var in obj.data
                    if not (z := var.data.loc[x]).empty
                ],
                obj.context,
            )

    @property
    def loc(self) -> _LocIndexer:
        return self._LocIndexer(self)

    @staticmethod
    def _antijoin(data: T, vars: Vars) -> T:
        return reduce(lambda d, v: d.pix.antijoin(v.data.index), vars.data, data)  # type: ignore

    def antijoin(self: SV, other: SV) -> SV:
        """
        Remove everything from self that is already in `other`

        Parameters
        ----------
        other : Vars
            Another set of derivations for the same variable

        Returns
        -------
        Vars
            Subset of self that is not already provided by `other`

        """
        return self.__class__(
            [
                z
                for var in self.data
                if not (z := Var(self._antijoin(var.data, other), var.provenance)).empty
            ],
            self.context,
        )

    def _missing(self, partial: bool | Var | MultiIndex = False) -> MultiIndex:
        full_index = self.context.full_index
        if isinstance(partial, Var):
            return full_index.join(
                partial.index(self.context.index), how="inner"
            ).pix.antijoin(partial.data.index)

        if isinstance(partial, MultiIndex):
            full_index = full_index.join(partial, how="inner")
        elif partial:
            full_index = full_index.join(self.index, how="inner")
        return self._antijoin(full_index, self)

    def missing(
        self, partial: bool = True, summarize: bool = True
    ) -> DataFrame | MultiIndex:
        index = self._missing(partial)
        return _summarize(index, self.context.index) if summarize else index

    def existing(self, summarize: bool = True) -> DataFrame | MultiIndex:
        index = concat(var.data.index for var in self.data)
        assert isinstance(index, MultiIndex)
        return _summarize(index, self.context.index) if summarize else index

    def _binop(self: SV, op: Callable[[Var, Var], Var], x: SV, y: SV) -> SV:
        if not all(isinstance(v, Vars) for v in (x, y)):
            return NotImplemented

        res = self.__class__(
            [z for u, v in product(x, y) if not (z := op(u, v)).empty], self.context
        )
        if self.context.optional_combinations:
            res = res | x | y
        return res

    def __add__(self: SV, other: SV) -> SV:
        if other == 0:
            return self
        return self._binop(operator.add, self, other)

    def __radd__(self: SV, other: SV) -> SV:
        if other == 0:
            return self
        return self._binop(operator.add, other, self)

    def __sub__(self: SV, other: SV) -> SV:
        if other == 0:
            return self
        return self._binop(operator.sub, self, other)

    def __rsub__(self: SV, other: SV) -> SV:
        if other == 0:
            return self
        return self._binop(operator.sub, other, self)

    def __mul__(self: SV, other: SV) -> SV:
        if other == 1:
            return self
        return self._binop(operator.mul, self, other)

    def __rmul__(self: SV, other: SV) -> SV:
        if other == 1:
            return self
        return self._binop(operator.mul, other, self)

    def __neg__(self: SV) -> SV:
        return self.__class__([-v for v in self], self.context)

    def __or__(self: SV, other: SV | float | int) -> SV:
        if isinstance(other, (float, int)):
            provenance = str(other)
            other_index = self._missing()
            if other_index.empty:
                return self
            other = DataFrame(
                other,
                index=other_index,
                columns=self.context.columns,
            )
            return self.__class__(self.data + [Var(other, provenance)], self.context)

        return self ^ other.antijoin(self)

    def __ror__(self: SV, other: SV) -> SV:
        return other ^ self.antijoin(other)

    def __xor__(self: SV, other: SV) -> SV:
        if not isinstance(other, Vars):
            return NotImplemented
        return self.__class__(self.data + other.data, self.context)

    def as_df(self, **assign: str) -> DataFrame:
        return concat(var.as_df(**assign) for var in self.data)


@define
class Resolver:
    """
    Resolver allows to consolidate variables by composing variants.

    Usage
    -----
    >>> co2emissions = ar6.loc[isin(gas="CO2")]
    >>> r = Resolver.from_data(co2emissions, "sector", ["AFOLU", "Energy"])
    >>> r["Energy"] |= r["Energy|Demand"] + r["Energy|Supply"]
    >>> r["AFOLU"] |= r["AFOLU|Land"] + r["AFOLU|Agriculture"]
    >>> r.as_df()
    """

    vars: dict[str, Vars]
    data: DataFrame
    context: Context  # context is shared with all Vars created
    external_data: dict[str, DataFrame]

    SR = TypeVar("SR", bound="Resolver")

    @classmethod
    def from_data(
        cls: type[SR],
        data: DataFrame,
        level: str,
        values: Sequence[str] | None = None,
        index: Sequence[str] = ("model", "scenario"),
        **external_data: DataFrame,
    ) -> SR:
        context = Context(
            level,
            full_index=data.index.droplevel(level).unique(),
            columns=data.columns,
            index=list(index),
        )
        inst = cls({}, data, context, external_data)
        if values is not None:
            for value in values:
                inst.add(value)
        return inst

    def add(self: SR, value: str, iamc_aggregate: bool = False) -> SR:
        vars = Vars.from_data(self.data, value, context=self.context)
        if iamc_aggregate:
            vars = vars | self.iamc_aggregate(value)
        self.vars[value] = vars
        return self

    class _LocIndexer:
        def __init__(self, obj: Resolver):
            self._obj = obj

        def __getitem__(self, x: Any) -> Resolver:
            obj = self._obj
            vars = {name: z for name, vars in obj.vars.items() if (z := vars.loc[x])}
            data = obj.data.loc[x]
            context = evolve(
                obj.context, full_index=data.index.droplevel(obj.context.level).unique()
            )
            return obj.__class__(vars, data, context, obj.external_data)

    @property
    def loc(self) -> _LocIndexer:
        return self._LocIndexer(self)

    def __len__(self) -> int:
        return len(self.vars)

    def __getitem__(self, value: str) -> Vars:
        vars = self.vars.get(value)
        if vars is not None:
            return vars

        vars = Vars.from_data(self.data, value, context=self.context)
        if not vars:
            try:
                # get variable from additional data
                prefix, rem_value = value.split("|", 1)
                data = self.external_data[prefix]
                if isinstance(data, Resolver):
                    vars = Vars.from_additionalresolver(
                        data[rem_value], prefix=prefix, context=self.context
                    )
                else:
                    vars = Vars.from_data(
                        data, rem_value, provenance=value, context=self.context
                    )
            except (KeyError, ValueError):
                vars = Vars([], context=self.context)

        return vars

    def _ipython_key_completions_(self) -> list[str]:
        comps = list(self.vars)
        comps.extend(self.data.pix.unique(self.context.level).difference(comps))  # type: ignore
        for n, v in self.external_data.items():
            comps.extend(f"{n}|" + v.pix.unique(self.context.level))  # type: ignore
        return comps

    @property
    def optional_combinations(self):
        return self.context.optional_combinations

    @property
    def index(self) -> MultiIndex:
        if not self.vars:
            return MultiIndex.from_tuples([], names=self.context.index)

        return reduce(
            MultiIndex.intersection, (vars.index for vars in self.vars.values())
        )

    def __repr__(self) -> str:
        lines = []
        num_scenarios = len(self.data.pix.unique(self.context.index))  # type: ignore
        level = self.context.level

        lines.append(
            f"Resolver with data for {num_scenarios} scenarios, "
            f"and {len(self)} defined {level}s for {len(self.index)} scenarios:"
        )
        lines.extend(
            f"* {name} ({len(vars)}): "
            + ", ".join(
                str(len(var.data.pix.unique(self.context.index)))  # type: ignore
                for var in vars.data
            )
            for name, vars in self.vars.items()
        )

        existing_provenances = set(
            v.provenance for vars in self.vars.values() for v in vars
        )
        unused_values = (
            self.data.pix.unique([*self.context.index, level])  # type: ignore
            .pix.antijoin(Index(self.vars, name=level).union(existing_provenances))
            .pix.project(level)
            .value_counts()
            .loc[lambda s: s > num_scenarios // 20]
        )
        lines.append(
            f"{len(unused_values)} {level}s for more than 5% of scenarios unused:"
        )
        lines.extend(f"* {value} ({num})" for value, num in unused_values.items())
        return "\n".join(lines)

    def __setitem__(self, value: str, vars: Vars) -> Vars:
        if not isinstance(vars, Vars):
            raise TypeError(f"Expected Vars instance, found: {type(vars)}")
        self.vars[value] = vars
        return vars

    def iamc_aggregate(self, value: str, **overwrites) -> Vars:
        pattern = f"{value}|*"
        overwritten_variables = {
            name: var
            for name, var in (self.vars | overwrites).items()
            if (
                re.match(shell_pattern_to_regex(pattern), name)
                and (var.data[0].provenance != name or len(var.data) > 1)
            )
        }

        data = (
            concat(
                [
                    self.data.loc[
                        ismatch(**{self.context.level: pattern})
                    ].pix.antijoin(
                        Index(overwritten_variables, name=self.context.level)
                    ),
                    *(
                        v.as_df(**{self.context.level: n}).droplevel("provenance")
                        for n, v in overwritten_variables.items()
                    ),
                ]
            )
            .groupby(self.data.index.names.difference([self.context.level]))
            .sum(min_count=1)
        )
        if data.empty:
            return Vars([], self.context)

        conditions = (
            (
                " with special "
                + ", ".join(name.removeprefix(value) for name in overwritten_variables)
            )
            if overwritten_variables
            else ""
        )
        provenance = f"sum({pattern}{conditions})"

        return Vars([Var(data, provenance)], self.context)

    def add_iamc_aggregate(self: SR, value: str) -> SR:
        self[value] |= self.iamc_aggregate(value)
        return self

    def as_df(self, only_consistent: bool = True) -> DataFrame:
        if only_consistent:
            index = self.index

            def maybe_consistent(df):
                return df.pix.semijoin(index, how="right")
        else:

            def maybe_consistent(df):
                return df

        return concat(
            vars.as_df(**{self.context.level: name}).pipe(maybe_consistent)
            for name, vars in self.vars.items()
        )
